/*******************************************************************************
* UART_problem
*
* S113 is present in the background, but not activated (project stripped from)

*
*******************************************************************************/

#define VERSION "0.01 / 20.01.2020"

#include "nordic_common.h"
#include "nrf.h"

#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"

#include <stdarg.h> // for var_args
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "boards.h"

#include "nrf_delay.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#include "nrf_pwr_mgmt.h"
#include "nrf_queue.h"

// Jo Toolbox and Libs
#include "tb_tools.h"
#include "ltx_defs.h"
#include "std_uart0.h"

#define APP_BLE_CONN_CFG_TAG 1                           /* tag identifying the SoftDevice BLE configuration. */

// -------------------------- Error Handlers -------------------------------
#define NRF_ERROR_OWN_BASE_NUM (0xF000) // Ab hier eigene Fehler (spaeter auslagern)
#define OWN_ERR_BUFFERSIZE NRF_ERROR_OWN_BASE_NUM + 1

#define LOC_ERROR_CHECK(res)            \
    {                                   \
        if (res != NRF_SUCCESS)         \
            local_error(res, __LINE__); \
    }

static void on_error(uint32_t err_code, uint32_t line_num, uint8_t* pfile){
    NRF_LOG_FINAL_FLUSH();
    tb_printf("FATAL ERROR: %X Line %u '%s'\n",err_code,line_num,pfile);
    nrf_delay_ms(200);
#ifdef DEBUG
    NRF_BREAKPOINT_COND;
#endif
    NVIC_SystemReset();
}
#define DEAD_BEEF 0xDEADBEEF /* Value used as error code on stack dump, can be used to identify stack location on stack unwind. */
void assert_nrf_callback(uint16_t line_num, const uint8_t *p_file_name) {
    NRF_LOG_ERROR("assert_nrf_callback()\n");
    NRF_LOG_FLUSH();
    on_error(DEAD_BEEF, line_num, (uint8_t *)p_file_name);
}
static void conn_params_error_handler(uint32_t nrf_error) {
    NRF_LOG_ERROR("conn_params_error_handler()\n");
    NRF_LOG_FLUSH();
    on_error(nrf_error, __LINE__, (uint8_t *)__FILE__);
}
static void local_error(uint32_t error, uint16_t line_num) {
    NRF_LOG_ERROR("ERROR %u in Line %u\n", error, line_num);
    NRF_LOG_FLUSH();
    on_error(error, line_num , (uint8_t *)__FILE__);
}

/**@brief Function for the SoftDevice initialization.
 *
 * @details This function initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void) {
    ret_code_t err_code;

    err_code = nrf_sdh_enable_request();
    LOC_ERROR_CHECK(err_code);

    // Configure the BLE stack using the default settings.
    // Fetch the start address of the application RAM.
    // 1 Connection braucht fast 4K RAM auf dem S113!
    uint32_t ram_start = 0, vorschlag;
    err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    LOC_ERROR_CHECK(err_code);

    // Enable BLE stack. - Das SD benoetigt eine gewisse Menge RAM, abstimmen mit App
#ifdef NRF_LOG
    vorschlag = ram_start;
#endif
    err_code = nrf_sdh_ble_enable(&ram_start);
#ifdef NRF_LOG
    NRF_LOG_INFO("Stack Frei: %d Bytes\n", (int32_t)(vorschlag - ram_start));
#endif
    LOC_ERROR_CHECK(err_code);

    err_code =sd_power_dcdc_mode_set(NRF_POWER_DCDC_ENABLE);
    LOC_ERROR_CHECK(err_code);
#if 0
    // Register a handler for BLE events.
    NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
#endif
}

/* Function for initializing the nrf log module. */
static void log_init(void) {
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    LOC_ERROR_CHECK(err_code);
    NRF_LOG_DEFAULT_BACKENDS_INIT();
}

/* Function for initializing power management. */
static void power_management_init(void) {
    ret_code_t err_code;
    err_code = nrf_pwr_mgmt_init();
    LOC_ERROR_CHECK(err_code);
}

/* Function for handling the idle state (main loop).
If there is no pending log operation, then sleep until next the next event occurs. */
static void idle_state_handle(void) {
    if (NRF_LOG_PROCESS() == false) 
    {
        nrf_pwr_mgmt_run();
    }
}


// Init UART and Toolbox
void jo_drv_init(void) {
    int16_t res;

    tb_init(); // Init Toolbox (ohne Watchdog) und Minimal (wie Std. pca10056)

    tb_watchdog_init();
    tb_watchdog_feed(1);

    tb_printf("\n\n*** UART_problem " VERSION " JoEmbedded.de\n\n");

#ifdef DEBUG
    uint32_t uval= (NRF_POWER->RESETREAS);
    (NRF_POWER->RESETREAS)=0xFFFFFFFF;  // Clear with '1'

    tb_printf("Reset-Reason: 0x%X ",uval);
    if(uval&1) tb_printf("(Pin-Reset)");  // The low Nibble Reasons
    if(uval&2) tb_printf("(Watchdog)");
    if(uval&4) tb_printf("(Soft-Reset)");
    if(uval&8) tb_printf("(CPU Lockup)");
    tb_printf("\n");
#endif
}


/************************** Lokale UART-Kommandos ***************************/
#define INPUT_LEN 80  
uint8_t input[INPUT_LEN + 1];
uint16_t input_len;


//=== cmdline ===
void uart_cmdline(void) {
    size_t cmdlen;
    uint32_t err_code;
    uint32_t val, val2;
    int16_t res;
    char *pc;

    tb_printf(">");
    cmdlen = tb_gets(input, INPUT_LEN, 15000, 1); // Max. 15 secs for input, echo
    tb_printf("\n");
    if (cmdlen) {
        pc = input + 1;
        val = strtol(pc, &pc, 0);  // Parameter 1
        val2 = strtol(pc, &pc, 0); // Parameter 2
        switch (*input) {
        // --------- Commands -------------------

        case 'R': 
            tb_printf("Reset...\n");
            tb_delay_ms(300);
            tb_system_reset();

        // Problem No 1: FPU Exception on sleep for large Numbers
        case 'x': // Enter as Input 'x1.71141e38' -> FPU Excpeption on powerdown
            tb_printf("Test: '%g'\n",atof(input+1));
            break;

        // Problem No 2: How to clear UART errors?
        case 'g':
            tb_printf("---GPS L76-L-- \n",val);
            //nrf_gpio_pin_clear(PIN_GPS_PWR);     // L: GPS Power ON
            //nrf_gpio_pin_set(PIN_GPS_RTCPWR);     // H: GPS Battery supply 
            //tb_delay_ms(10);
            stu_init();
            val=0;

            for(;;){

                tb_watchdog_feed(1);

                // Input from stu_uart?
                res=stu_kbhit();
                if(res<0) {
                  tb_printf("ERROR: %d - stu_kbhit()\n",res);
                }else if(res){  // OK: Input from stu_uart:
                   res=stu_getc();
                   if(res<0){
                    tb_printf("ERROR: %d - stu_getc()\n",res);
                   }else{
                    tb_putc(res);
                    }
                }
          
                if(res<0){
                    tb_printf("STU: Statistic: EDC:%u EFC:%u RX:%u TX:%u\n",
                        stu_stat.drv_error_cnt,stu_stat.fifo_error_cnt, stu_stat.rx_data_cnt, stu_stat.tx_done_cnt);
                    tb_delay_ms(1000);
                }
              
                // user pressed key?
                if(tb_kbhit()){
                  val=tb_getc();
                  if(val=='w'){
                    tb_printf("\n<WAIT 10 sec>...");
                    tb_delay_ms(10000);
                    tb_printf("<OK>\n");
                  }else if(val=='q' || val=='Q'){
                    break;  // exit with Power ON or OFF

                  }else if(val=='m'){
                    tb_printf("\n<MINIMUM RMC>\n"); // Set to Minimum Output RMC
                    stu_printf("$PMTK314,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29\r\n");
                  }else if(val=='b' ){ 
                    tb_printf("\n<STDBYE>\n"); // GPS to Standby 
                    stu_printf("$PMTK225,4*2F\r\n");
                  }else if(val=='c'){
                    tb_printf("\n<CLEAR ERROR STATE>\n"); 
                    stu_clear_error_state();
                  }
                }
            }

            stu_uninit();
            if(val=='Q'){
              tb_printf("---GPS off---\n");
              // // Keep RTC of GPS running  --- nrf_gpio_pin_clear(PIN_GPS_RTCPWR);     // L: Battery Off
              //nrf_gpio_pin_set(PIN_GPS_PWR);      // H: Off
            }else{
              tb_printf("---keep on GPS---\n");
            }
            break;

        default:
            tb_printf("???\n");
        }
    } // if (cmdlen)

}


/***************** M A I N *************************/
 int main(void) {

    log_init();

    power_management_init();
    jo_drv_init();    //tb_tools etc
    ble_stack_init(); // Only to enbale clocks etc.

    NRF_LOG_INFO("Minimum Debug logging RTT started.\n");
#ifdef DEBUG
    tb_printf("*** DEBUG - " __DATE__ " "__TIME__ " ***\n");
    if(tb_is_wd_init()) tb_printf("INFO: Watchdog is ON!\n");
#endif

    // Enter main loop.
    for (;;) {

        tb_watchdog_feed(1);
        
        // ---- UART-Service ---
        if (tb_kbhit()) {
            uart_cmdline();
        }

        idle_state_handle();
    }
}
// ***