/*******************************************************************************
* STD_UART0.C - Lib for UART0
*
* STD_UART send data in small blaocks (later: Software HW-Flow)
*
* First Test with L76-L GPS
*
*******************************************************************************/

#include "nordic_common.h"
#include "nrf.h"  // Special CPU/SDK Defs z.B. NRF52_SERIES, 'nrf52840.h'

#include <stdarg.h> // for var_args
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "boards.h"
#include "app_uart.h"
#include "nrf_delay.h"
#include "nrf_drv_twi.h"
#include "nrf_serial.h" // components/libraries/serial


// JW Toolbox
#include "tb_tools.h"

#include "ltx_defs.h"
#include "std_uart0.h"

#if defined (UARTE_PRESENT) // For Test-Fkts
  #include "nrf_uarte.h"
#endif

#define STD_UART_NO  0 // use UARTE(0)

#define STD_UART_TX_PIN		NRF_GPIO_PIN_MAP(1,10) 		// GPS
#define STD_UART_RX_PIN		NRF_GPIO_PIN_MAP(1,6) 		// GPS
#define STD_UART_CTS_PIN	NRF_UART_PSEL_DISCONNECTED // Optional
#define STD_UART_RTS_PIN	NRF_UART_PSEL_DISCONNECTED // Optional
#define STD_UART_BAUDRATE	NRF_UART_BAUDRATE_9600


NRF_SERIAL_UART_DEF(std_uart,STD_UART_NO); // reserve Name+InterfaceNumber
NRF_SERIAL_DRV_UART_CONFIG_DEF(m_std_uart_drv_config,
                      STD_UART_RX_PIN, STD_UART_TX_PIN,
                      NRF_UART_PSEL_DISCONNECTED /*RTS_PIN_NUMBER*/, NRF_UART_PSEL_DISCONNECTED /*CTS_PIN_NUMBER*/,
                      NRF_UART_HWFC_ENABLED, NRF_UART_PARITY_EXCLUDED,
                      STD_UART_BAUDRATE,
                      UART_DEFAULT_CONFIG_IRQ_PRIORITY);

// UART requires FIFOs and working Buffers
#define SERIAL_FIFO_TX_SIZE 16	// Small TX Buffer to Monitor HW-Flow
#define SERIAL_FIFO_RX_SIZE 512 // Larger input buffer
NRF_SERIAL_QUEUES_DEF(std_uart_queues, SERIAL_FIFO_TX_SIZE, SERIAL_FIFO_RX_SIZE);

#define SERIAL_BUFF_TX_SIZE 8 // TX Junks: Blocks: >=1 
#define SERIAL_BUFF_RX_SIZE 1 // RX in single Chars: 1
NRF_SERIAL_BUFFERS_DEF(std_uart_buffs, SERIAL_BUFF_TX_SIZE, SERIAL_BUFF_RX_SIZE);

// Declare Fwd.
void std_uart_handler(struct nrf_serial_s const *p_serial, nrf_serial_event_t event);

NRF_SERIAL_CONFIG_DEF(std_uart_config, 
                    NRF_SERIAL_MODE_DMA, //  Mode: Polling/IRQ/DMA (UART 1 requires DMA)
                    &std_uart_queues, 
                    &std_uart_buffs, 
                    std_uart_handler, // No event handler	(later for Flow)
                    NULL); // No sleep handler

#define STU_SIZE_LINE_OUT  128  // 1 terminal line working buffer 
static char stu_uart_line_out[STU_SIZE_LINE_OUT];

static bool stu_init_flag=false; 

#define STU_STATE_EVTDRV_ERROR  1 
#define STU_STATE_FIFO_ERROR  2

STU_STATISTICS stu_stat;

//---------- init ---------------------
void stu_init(void){
    ret_code_t ret;
    if(stu_init_flag==false){  // Higher Power Peripherals, can be powered off by uninit
      memset(&stu_stat,0,sizeof(stu_stat));  // Clear statistics
      ret = nrf_serial_init(&std_uart, &m_std_uart_drv_config, &std_uart_config);
      APP_ERROR_CHECK(ret);
      stu_init_flag=true;
    }

}
// ------ uninit all higher power peripherals  --------------------
void stu_uninit(void){
    ret_code_t ret;

    // uninit higher power blocks
    if(stu_init_flag==true){    
      nrf_drv_uart_rx_abort(&std_uart.instance);
      ret=nrf_serial_flush(&std_uart, NRF_SERIAL_MAX_TIMEOUT); 
      APP_ERROR_CHECK(ret);
      ret = nrf_serial_uninit(&std_uart);
      APP_ERROR_CHECK(ret);

      // Strange Error, UART needs Power-Cycle for DeepSleep (nrf52840)
      // ( https://devzone.nordicsemi.com/f/nordic-q-a/54696/increased-current-consumption-after-nrf_serial_uninit )
#if std_uart_NO==0
      #define std_uart_BASE NRF_UARTE0_BASE  // 0x40002000
#elif std_uart_NO==1
      #define std_uart_BASE  NRF_UARTE1_BASE  // 0x40028000
#endif
     *(volatile uint32_t *)(std_uart_BASE + 0xFFC) = 0;
     *(volatile uint32_t *)(std_uart_BASE + 0xFFC);
     *(volatile uint32_t *)(std_uart_BASE + 0xFFC) = 1;
      stu_init_flag=false;
    }
}

// Recover from Errors (Problem? Does not work)
void stu_clear_error_state(void){
  ret_code_t ret;
  stu_stat.error_state=0;
  ret=nrf_serial_rx_drain(&std_uart);
  APP_ERROR_CHECK(ret);
}

// IRQ-handler is in 'nrfx_uarte.c' ca. Line 640, UART0
void std_uart_handler(struct nrf_serial_s const *p_serial, nrf_serial_event_t event){
  switch(event){
    case NRF_SERIAL_EVENT_DRV_ERR:
          stu_stat.error_state|=STU_STATE_EVTDRV_ERROR;
#ifdef DEBUG
          stu_stat.drv_error_cnt++;
#endif
          break;
    case NRF_SERIAL_EVENT_FIFO_ERR:
          stu_stat.error_state|=STU_STATE_FIFO_ERROR;
#ifdef DEBUG
          stu_stat.fifo_error_cnt++;
#endif
          break;
    case NRF_SERIAL_EVENT_RX_DATA: 
#ifdef DEBUG
          stu_stat.rx_data_cnt++;
#endif
          break;
    case NRF_SERIAL_EVENT_TX_DONE:  // Can be used for Manual HW-Flow
#ifdef DEBUG
          stu_stat.tx_done_cnt++;
#endif
          break;
    default:
        break;
    }
}

// Send Binary Data Block, wait until place in Queue. t.b.d: Software HW-Flow
int16_t stu_tx_bin_block(uint8_t *pu, size_t ulen){
    ret_code_t ret;
    size_t uavail;
	while(ulen){
		for(;;){
			// get available space in TX queue
			uavail=nrf_queue_available_get(&std_uart_queues_txq);
			if(uavail>=SERIAL_BUFF_TX_SIZE) break;  // enough space: OK to send
			tb_delay_ms(1);
		}
		if(uavail>ulen) uavail=ulen;
		ret = nrf_serial_write(&std_uart,pu,uavail, NULL,0); 
		APP_ERROR_CHECK(ret);  // NRF_SUCCESS if OK
		
		ulen-=uavail;
		pu+=uavail;
	}
	return 0; // OK
}


// stu_putc(): Wait if not ready. Return <0 on Failure or 0
int16_t stu_putc(char c){
    if(stu_init_flag==false) return -ERROR_NOT_INIT; // Not init...
	return stu_tx_bin_block((uint8_t*)&c,1);
}

int16_t stu_printf(char* fmt, ...){
    ret_code_t ret;
    size_t ulen;
    va_list argptr;
    va_start(argptr, fmt);

    if(stu_init_flag==false) return -ERROR_NOT_INIT; // Not init...

    ulen=vsnprintf((char*)stu_uart_line_out, STU_SIZE_LINE_OUT, fmt, argptr);  // vsn: limit!
    va_end(argptr);
    // vsnprintf() limits output to given size, but might return more.
    if(ulen>STU_SIZE_LINE_OUT-1) return -ERROR_TOO_MUCH_DATA;
	return stu_tx_bin_block(stu_uart_line_out,ulen);
}


// ---- Input functions 0: Nothing available (stu_kbhit is faster than stu_getc) ---------
int16_t stu_kbhit(void){
    int16_t res;
    if(stu_init_flag==false) return -ERROR_NOT_INIT; // Not init...
    if(stu_stat.error_state) return -(ERROR_DRV_ERROR-1+stu_stat.error_state);

    res=!nrf_queue_is_empty(&std_uart_queues_rxq);
    return res;
}

// ---- get 1 char (0..255) (or -1 if nothing available)
int16_t stu_getc(void){
    ret_code_t ret;
    uint8_t c;

    if(stu_init_flag==false) return -ERROR_NOT_INIT; // Not init...
    if(stu_stat.error_state) return -(ERROR_DRV_ERROR-1+stu_stat.error_state);

    ret=nrf_queue_generic_pop(&std_uart_queues_rxq,&c,false);
    if(ret!=NRF_SUCCESS) return -1;
    return (int16_t)c;  
}


// Get String with Timout (if >0) in msec of infinite (Timout 0), No Echo, Noch BS for Sensors!
int16_t stu_gets(char* input, int16_t max_uart_in, uint16_t max_wait_ms, uint8_t flags){
    int16_t idx=0;
    int16_t res;
    char c;
    max_uart_in--;  // Reserve 1 Byte for End-0

    if(stu_init_flag==false) return -ERROR_NOT_INIT; // Not init...

    for(;;){
        res=stu_kbhit();
        if(res>0){
            res=stu_getc();
            if(res<0) break;
            c=res;
            if(c=='\n' || c=='\r') {  // Kein Backspace (8) bei Sensoren
                if(c=='\r'){
                  if(flags & STU_IGNORE_CR) continue;
                }else{
                  if(flags & STU_IGNORE_LF) continue;
                }
                break;    // NL CR or whatever (no Echo for NL CR)
            }else if(c>=' ' && c<128 && idx<max_uart_in){
                input[idx++]=c;
            }
        }else if(!res){
            if(max_wait_ms){
                if(!--max_wait_ms){
                  idx=0;  // Return empty String
                  break;
                }
            }
            tb_delay_ms(1);
        }else return res; // ERROR
    }
    input[idx]=0;   // Terminate String
    return idx;
}

//**
